
#include <stdio.h>
#include <stdlib.h>

#define		FALSE	0
#define		TRUE	1

void show_usage( void )
{
  printf( "usage: gcaback [-v] [FILE]...\n" );
  printf( "       -v: verbose mode\n" );
}

int main( int argc, char * argv[] )
{
  int option_valid;
  int option_verbose;
  int filegiven;
  FILE *inputfile;
  FILE *outputfile;
  int gca_id_found;
  int filename_length;
  char filename[ FILENAME_MAX + 2 ] ;
  int embedded_file_length;
  int file_nr;

  // at least two arguments?
  if ( argc < 2 ) 
  {
    show_usage();
    return EXIT_FAILURE;
  }

  // set no file given
  filegiven = FALSE;

  // default values for options
  option_verbose = FALSE;

  // process all given files
  for( file_nr = 1; file_nr < argc; file_nr++ )
  {
    // disable option validity
    option_valid = FALSE;

    // check verbose mode (option -v)
    if ( !strncmp( argv[ file_nr ], "-v", 2 ) )
    {
      option_verbose = TRUE;
      option_valid = TRUE;
    }

    // file opening successful?
    if ( !option_valid )
    {
      filegiven = TRUE;

      if ( inputfile = fopen( argv[ file_nr ], "rb" ) )
      {
        int end_of_file = FALSE;
        char r_char;
        int gca_id_state = 0;

        if ( option_verbose )
          printf( "opened file %s\n", argv[ file_nr ] );

        // search for 'GCA' ID
        gca_id_found = FALSE;

        while ( end_of_file == FALSE )
        {
          r_char = fgetc( inputfile );

          if ( feof( inputfile ) )
            end_of_file = TRUE;
          else
          {
            switch( gca_id_state )
            {
              // nothing found yet
              case 0:
                if ( r_char == 'G' )
                  gca_id_state = 1;
                else
                  gca_id_state = 0;
                break;

              // 'G' found
              case 1:
                if ( r_char == 'C' )
                  gca_id_state = 2;
                else
                  gca_id_state = 0;
                break;

              // 'GC' found
              case 2:
                if ( r_char == 'A' )
                {
                  int end_of_filename = FALSE;
                  filename_length = 0;

                  gca_id_found = TRUE;

                  if ( option_verbose )
                    printf( "GCA ID found\n" );

                  // get name of embedded file
                  while ( ( end_of_filename == FALSE ) && ( filename_length < ( FILENAME_MAX + 1 ) ) )
                  {
                    r_char = fgetc( inputfile );

                    if ( feof( inputfile ) )
                    {
                      if ( !option_verbose )
                        printf( "%s: ", argv[ file_nr ] );
                      printf( "unexpected end of file\n" );
                      fclose( inputfile );
                      return EXIT_FAILURE;
                    }

                    if ( r_char == '\0' )
                    {
                      filename[ filename_length ] = '\0';
                      end_of_filename = TRUE;
                    }
                    else if ( r_char >= 32 )
                    {
                      filename[ filename_length ] = r_char;
                      filename_length++;
                    }
                    else
                    {
                      if ( !option_verbose )
                        printf( "%s: ", argv[ file_nr ] );
                      printf( "strange embedded filename... are you sure this is a GCA file?\n" );
                      fclose( inputfile );
                      return EXIT_FAILURE;
                    }
                  }

                  // length of filename ok?
                  if ( filename_length >= FILENAME_MAX + 1 )
                  {
                    if ( !option_verbose )
                      printf( "%s: ", argv[ file_nr ] );
                    printf( "filename of embedded file too long\n" );
                    fclose( inputfile );
                    return EXIT_FAILURE;
                  }

                  if ( option_verbose )
                    printf( "embedded file name: %s\n", filename );

                  // get length of embedded file (little endian->big endian)
                  embedded_file_length  = ( fgetc( inputfile ) << 24 );
                  embedded_file_length += ( fgetc( inputfile ) << 16 );
                  embedded_file_length += ( fgetc( inputfile ) <<  8 );
                  embedded_file_length +=   fgetc( inputfile );

                  if ( option_verbose )
                    printf( "embedded file length: %d\n", embedded_file_length );

                  // extract embedded file
                  if ( outputfile = fopen( filename, "wb" ) )
                  {
                    int counter;

                    if ( option_verbose )
                      printf( "extracting embedded file...\n" );

                    for ( counter = 0; counter < embedded_file_length; counter++ )
                    {
                      r_char = fgetc( inputfile );

                      if ( feof( inputfile ) )
                      {
                        if ( !option_verbose )
                          printf( "%s: ", argv[ file_nr ] );
                        printf( "unexpected end of file\n" );
                        fclose( inputfile );
                        fclose( outputfile );
                        return EXIT_FAILURE;
                      }

                      fputc( r_char, outputfile );
                    }

                    fclose( outputfile );
                  }

                  if ( option_verbose )
                    printf( "done!\n" );
                }
                else
                  gca_id_state = 0;
                break;

              default:
                break;
            }
          }
        }

        if ( gca_id_found == FALSE )
        {
          if ( !option_verbose )
            printf( "%s: ", argv[ file_nr ] );
          printf( "no GCA ID found\n" );
        }

        fclose( inputfile );
      }
      else
        printf( "%s: file not found\n", argv[ file_nr ] );
    }
  }

  if ( filegiven == FALSE ) 
  {
    show_usage();
    return EXIT_FAILURE;
  }

  return EXIT_SUCCESS;
}



